"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const getWorkspace_1 = require("../workspace/getWorkspace");
const workerQueue_1 = require("../task/workerQueue");
const TaskLogger_1 = require("../logger/TaskLogger");
const backfill_1 = require("../cache/backfill");
const Pipeline_1 = require("../task/Pipeline");
const taskId_1 = require("../task/taskId");
// Run multiple
async function worker(cwd, config, reporters) {
    const workspace = getWorkspace_1.getWorkspace(cwd, config);
    const pipeline = new Pipeline_1.Pipeline(workspace, config);
    const { workerQueue, redisClient } = await workerQueue_1.initWorkerQueue(config.workerQueueOptions);
    const pubSubListener = (channel, message) => {
        if (workerQueue_1.workerPubSubChannel === channel) {
            if (message === "done") {
                workerQueue.close();
                redisClient.off("message", pubSubListener);
                redisClient.unsubscribe(workerQueue_1.workerPubSubChannel);
                redisClient.quit();
            }
        }
    };
    redisClient.subscribe(workerQueue_1.workerPubSubChannel);
    redisClient.on("message", pubSubListener);
    workerQueue.ready(() => {
        workerQueue.process(config.concurrency, async (job) => {
            const id = job.data.id;
            if (!pipeline.targets.has(id)) {
                return;
            }
            const target = pipeline.targets.get(id);
            const deps = getDepsForTarget(id, pipeline.dependencies);
            const logger = new TaskLogger_1.TaskLogger(job.data.name, job.data.task);
            logger.info(`processing job ${job.id}`);
            await Promise.all(deps.filter(d => d !== Pipeline_1.START_TARGET_ID).map((depTargetId) => {
                return getCache(pipeline.targets.get(depTargetId), workspace.root, config);
            }));
            const cacheResult = await getCache(target, workspace.root, config);
            if (cacheResult.cacheHit) {
                logger.info(`skipped ${id}`);
                return;
            }
            let result;
            if (target.packageName) {
                result = target.run({
                    packageName: target.packageName,
                    config,
                    cwd: target.cwd,
                    options: target.options,
                    taskName: taskId_1.getPackageAndTask(target.id).task,
                    logger,
                });
            }
            else {
                result = target.run({
                    config,
                    cwd: target.cwd,
                    options: target.options,
                    logger,
                });
            }
            if (result && typeof result["then"] === "function") {
                await result;
            }
            await saveCache(cacheResult.hash, target, config);
        });
    });
}
exports.worker = worker;
// speeding up to reduce network costs for a worker
const localHashCache = {};
function getCacheOptions(target, config) {
    return Object.assign(Object.assign({}, config.cacheOptions), { outputGlob: target.outputGlob || config.cacheOptions.outputGlob });
}
async function getCache(target, root, config) {
    let hash = null;
    let cacheHit = false;
    const { id, cwd } = target;
    if (localHashCache[id]) {
        return { hash: localHashCache[id], cacheHit: true };
    }
    const cacheOptions = getCacheOptions(target, config);
    hash = await backfill_1.cacheHash(id, root, cwd, cacheOptions, config.args);
    if (hash && !config.resetCache) {
        cacheHit = await backfill_1.cacheFetch(hash, id, cwd, config.cacheOptions);
        if (cacheHit) {
            localHashCache[id] = hash;
        }
    }
    return { hash, cacheHit };
}
async function saveCache(hash, target, config) {
    const localCacheKey = target.id;
    localHashCache[localCacheKey] = hash;
    const cacheOptions = getCacheOptions(target, config);
    await backfill_1.cachePut(hash, target.cwd, cacheOptions);
}
function getDepsForTarget(id, dependencies) {
    const stack = [id];
    const deps = new Set();
    const visited = new Set();
    while (stack.length > 0) {
        const current = stack.pop();
        if (visited.has(current)) {
            continue;
        }
        visited.add(current);
        if (current !== id) {
            deps.add(current);
        }
        dependencies.forEach(([from, to]) => {
            if (to === current) {
                stack.push(from);
            }
        });
    }
    return [...deps];
}
