"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const TaskLogger_1 = require("../logger/TaskLogger");
const abortSignal_1 = require("./abortSignal");
const formatDuration_1 = require("../logger/reporters/formatDuration");
class DistributedTarget {
    constructor(target, config, context, workerQueue) {
        this.target = target;
        this.config = config;
        this.context = context;
        this.workerQueue = workerQueue;
        this.startTime = [0, 0];
        this.duration = [0, 0];
        this.status = "pending";
        this.logger = new TaskLogger_1.TaskLogger(target.packageName || "[GLOBAL]", target.packageName ? target.task : target.id);
        this.cacheOptions = Object.assign(Object.assign({}, config.cacheOptions), { outputGlob: [...(config.cacheOptions.outputGlob || []), ...(target.outputGlob || [])] });
        this.context.targets.set(target.id, this);
    }
    onStart() {
        this.status = "started";
        this.startTime = process.hrtime();
        this.logger.info("started", { status: "started" });
    }
    onComplete() {
        this.status = "completed";
        this.duration = process.hrtime(this.startTime);
        this.logger.info("completed", {
            status: "completed",
            duration: formatDuration_1.hrToSeconds(this.duration),
        });
    }
    onFail() {
        this.status = "failed";
        this.duration = process.hrtime(this.startTime);
        this.logger.info("failed", {
            status: "failed",
            duration: formatDuration_1.hrToSeconds(this.duration),
        });
    }
    onSkipped(hash) {
        this.status = "skipped";
        this.duration = process.hrtime(this.startTime);
        this.logger.info(`skipped`, {
            status: "skipped",
            duration: formatDuration_1.hrToSeconds(this.duration),
            hash,
        });
    }
    async run() {
        const { target, context, config, logger } = this;
        try {
            this.onStart();
            // Wraps with profiler as well as task args
            await context.profiler.run(() => {
                return new Promise((resolve, reject) => {
                    const job = this.workerQueue.createJob({
                        id: target.id,
                    });
                    job.on("succeeded", (result) => {
                        logger.info("succeeded");
                        resolve();
                    });
                    job.on("failed", (result) => {
                        logger.info("failed");
                        reject();
                    });
                    // times out at 1 hour
                    job
                        .timeout(1000 * 60 * 60)
                        .save()
                        .then((newJob) => {
                        logger.info(`job id: ${newJob.id}`);
                    });
                });
            }, target.id);
            this.onComplete();
        }
        catch (e) {
            context.measures.failedTargets = context.measures.failedTargets || [];
            context.measures.failedTargets.push(target.id);
            this.onFail();
            if (config.continue) {
                return true;
            }
            if (!config.safeExit) {
                abortSignal_1.controller.abort();
            }
            return false;
        }
        return true;
    }
}
exports.DistributedTarget = DistributedTarget;
